@@ -7,8 +7,8 @@ from unittest import mock
 import pytest
 
 import aiohttp
-from aiohttp import ClientWebSocketResponse, errors, hdrs, helpers
-from aiohttp._ws_impl import WS_KEY
+from aiohttp import client, hdrs, helpers
+from aiohttp.http import WS_KEY
 from aiohttp.log import ws_logger
 
 
@@ -43,12 +43,11 @@ def test_ws_connect(ws_key, loop, key_data):
             m_req.return_value = helpers.create_future(loop)
             m_req.return_value.set_result(resp)
 
-            res = yield from aiohttp.ws_connect(
+            res = yield from aiohttp.ClientSession(loop=loop).ws_connect(
                 'http://test.org',
-                protocols=('t1', 't2', 'chat'),
-                loop=loop)
+                protocols=('t1', 't2', 'chat'))
 
-    assert isinstance(res, ClientWebSocketResponse)
+    assert isinstance(res, client.ClientWebSocketResponse)
     assert res.protocol == 'chat'
     assert hdrs.ORIGIN not in m_req.call_args[1]["headers"]
 
@@ -64,10 +63,9 @@ def test_ws_connect_with_origin(key_data, loop):
             m_req.return_value.set_result(resp)
 
             origin = 'https://example.org/page.html'
-            with pytest.raises(errors.WSServerHandshakeError):
-                yield from aiohttp.ws_connect('http://test.org',
-                                              loop=loop,
-                                              origin=origin)
+            with pytest.raises(client.WSServerHandshakeError):
+                yield from aiohttp.ClientSession(loop=loop).ws_connect(
+                    'http://test.org', origin=origin)
 
     assert hdrs.ORIGIN in m_req.call_args[1]["headers"]
     assert m_req.call_args[1]["headers"][hdrs.ORIGIN] == origin
@@ -76,7 +74,7 @@ def test_ws_connect_with_origin(key_data, loop):
 @asyncio.coroutine
 def test_ws_connect_custom_response(loop, ws_key, key_data):
 
-    class CustomResponse(ClientWebSocketResponse):
+    class CustomResponse(client.ClientWebSocketResponse):
         def read(self, decode=False):
             return 'customized!'
 
@@ -93,37 +91,13 @@ def test_ws_connect_custom_response(loop, ws_key, key_data):
             m_req.return_value = helpers.create_future(loop)
             m_req.return_value.set_result(resp)
 
-            res = yield from aiohttp.ws_connect(
-                'http://test.org',
-                ws_response_class=CustomResponse,
-                loop=loop)
+            res = yield from aiohttp.ClientSession(
+                ws_response_class=CustomResponse, loop=loop).ws_connect(
+                    'http://test.org')
 
     assert res.read() == 'customized!'
 
 
-@asyncio.coroutine
-def test_ws_connect_global_loop(loop, ws_key, key_data):
-    asyncio.set_event_loop(loop)
-
-    resp = mock.Mock()
-    resp.status = 101
-    resp.headers = {
-        hdrs.UPGRADE: hdrs.WEBSOCKET,
-        hdrs.CONNECTION: hdrs.UPGRADE,
-        hdrs.SEC_WEBSOCKET_ACCEPT: ws_key
-    }
-    with mock.patch('aiohttp.client.os') as m_os:
-        with mock.patch('aiohttp.client.ClientSession.get') as m_req:
-            m_os.urandom.return_value = key_data
-            m_req.return_value = helpers.create_future(loop)
-            m_req.return_value.set_result(resp)
-
-            resp = yield from aiohttp.ws_connect('http://test.org')
-    assert resp._loop is loop
-
-    asyncio.set_event_loop(None)
-
-
 @asyncio.coroutine
 def test_ws_connect_err_status(loop, ws_key, key_data):
     resp = mock.Mock()
@@ -139,10 +113,10 @@ def test_ws_connect_err_status(loop, ws_key, key_data):
             m_req.return_value = helpers.create_future(loop)
             m_req.return_value.set_result(resp)
 
-            with pytest.raises(errors.WSServerHandshakeError) as ctx:
-                yield from aiohttp.ws_connect('http://test.org',
-                                              protocols=('t1', 't2', 'chat'),
-                                              loop=loop)
+            with pytest.raises(client.WSServerHandshakeError) as ctx:
+                yield from aiohttp.ClientSession(loop=loop).ws_connect(
+                    'http://test.org',
+                    protocols=('t1', 't2', 'chat'))
 
     assert ctx.value.message == 'Invalid response status'
 
@@ -162,10 +136,11 @@ def test_ws_connect_err_upgrade(loop, ws_key, key_data):
             m_req.return_value = helpers.create_future(loop)
             m_req.return_value.set_result(resp)
 
-            with pytest.raises(errors.WSServerHandshakeError) as ctx:
-                yield from aiohttp.ws_connect('http://test.org',
-                                              protocols=('t1', 't2', 'chat'),
-                                              loop=loop)
+            with pytest.raises(client.WSServerHandshakeError) as ctx:
+                yield from aiohttp.ClientSession(loop=loop).ws_connect(
+                    'http://test.org',
+                    protocols=('t1', 't2', 'chat'))
+
     assert ctx.value.message == 'Invalid upgrade header'
 
 
@@ -184,10 +159,10 @@ def test_ws_connect_err_conn(loop, ws_key, key_data):
             m_req.return_value = helpers.create_future(loop)
             m_req.return_value.set_result(resp)
 
-            with pytest.raises(errors.WSServerHandshakeError) as ctx:
-                yield from aiohttp.ws_connect('http://test.org',
-                                              protocols=('t1', 't2', 'chat'),
-                                              loop=loop)
+            with pytest.raises(client.WSServerHandshakeError) as ctx:
+                yield from aiohttp.ClientSession(loop=loop).ws_connect(
+                    'http://test.org',
+                    protocols=('t1', 't2', 'chat'))
 
     assert ctx.value.message == 'Invalid connection header'
 
@@ -207,10 +182,10 @@ def test_ws_connect_err_challenge(loop, ws_key, key_data):
             m_req.return_value = helpers.create_future(loop)
             m_req.return_value.set_result(resp)
 
-            with pytest.raises(errors.WSServerHandshakeError) as ctx:
-                yield from aiohttp.ws_connect('http://test.org',
-                                              protocols=('t1', 't2', 'chat'),
-                                              loop=loop)
+            with pytest.raises(client.WSServerHandshakeError) as ctx:
+                yield from aiohttp.ClientSession(loop=loop).ws_connect(
+                    'http://test.org',
+                    protocols=('t1', 't2', 'chat'))
 
     assert ctx.value.message == 'Invalid challenge response'
 
@@ -251,7 +226,7 @@ def test_ws_connect_common_headers(ws_key, loop, key_data):
                     protocols=('t1', 't2', 'chat'),
                     headers=headers)
 
-        assert isinstance(res, ClientWebSocketResponse)
+        assert isinstance(res, client.ClientWebSocketResponse)
         assert res.protocol == 'chat'
         assert hdrs.ORIGIN not in m_req.call_args[1]["headers"]
 
@@ -277,16 +252,14 @@ def test_close(loop, ws_key, key_data):
                 m_req.return_value = helpers.create_future(loop)
                 m_req.return_value.set_result(resp)
                 writer = WebSocketWriter.return_value = mock.Mock()
-                reader = mock.Mock()
-                resp.connection.reader.set_parser.return_value = reader
 
-                resp = yield from aiohttp.ws_connect('http://test.org',
-                                                     loop=loop)
+                session = aiohttp.ClientSession(loop=loop)
+                resp = yield from session.ws_connect(
+                    'http://test.org')
                 assert not resp.closed
 
-                msg = aiohttp.WSMessage(aiohttp.MsgType.CLOSE, b'', b'')
-                reader.read.return_value = helpers.create_future(loop)
-                reader.read.return_value.set_result(msg)
+                resp._reader.feed_data(
+                    aiohttp.WSMessage(aiohttp.WSMsgType.CLOSE, b'', b''), 0)
 
                 res = yield from resp.close()
                 writer.close.assert_called_with(1000, b'')
@@ -299,6 +272,8 @@ def test_close(loop, ws_key, key_data):
                 assert not res
                 assert writer.close.call_count == 1
 
+                session.close()
+
 
 @asyncio.coroutine
 def test_close_exc(loop, ws_key, key_data):
@@ -316,21 +291,20 @@ def test_close_exc(loop, ws_key, key_data):
                 m_req.return_value = helpers.create_future(loop)
                 m_req.return_value.set_result(resp)
                 WebSocketWriter.return_value = mock.Mock()
-                reader = mock.Mock()
-                resp.connection.reader.set_parser.return_value = reader
 
-                resp = yield from aiohttp.ws_connect('http://test.org',
-                                                     loop=loop)
+                session = aiohttp.ClientSession(loop=loop)
+                resp = yield from session.ws_connect('http://test.org')
                 assert not resp.closed
 
                 exc = ValueError()
-                reader.read.return_value = helpers.create_future(loop)
-                reader.read.return_value.set_exception(exc)
+                resp._reader.set_exception(exc)
 
                 yield from resp.close()
                 assert resp.closed
                 assert resp.exception() is exc
 
+                session.close()
+
 
 @asyncio.coroutine
 def test_close_exc2(loop, ws_key, key_data):
@@ -348,10 +322,9 @@ def test_close_exc2(loop, ws_key, key_data):
                 m_req.return_value = helpers.create_future(loop)
                 m_req.return_value.set_result(resp)
                 writer = WebSocketWriter.return_value = mock.Mock()
-                resp.connection.reader.set_parser.return_value = mock.Mock()
 
-                resp = yield from aiohttp.ws_connect('http://test.org',
-                                                     loop=loop)
+                resp = yield from aiohttp.ClientSession(loop=loop).ws_connect(
+                    'http://test.org')
                 assert not resp.closed
 
                 exc = ValueError()
@@ -382,8 +355,8 @@ def test_send_data_after_close(ws_key, key_data, loop, mocker):
             m_req.return_value = helpers.create_future(loop)
             m_req.return_value.set_result(resp)
 
-            resp = yield from aiohttp.ws_connect('http://test.org',
-                                                 loop=loop)
+            resp = yield from aiohttp.ClientSession(loop=loop).ws_connect(
+                'http://test.org')
             resp._writer._closing = True
 
             mocker.spy(ws_logger, 'warning')
@@ -415,8 +388,8 @@ def test_send_data_type_errors(ws_key, key_data, loop):
                 m_req.return_value.set_result(resp)
                 WebSocketWriter.return_value = mock.Mock()
 
-                resp = yield from aiohttp.ws_connect('http://test.org',
-                                                     loop=loop)
+                resp = yield from aiohttp.ClientSession(loop=loop).ws_connect(
+                    'http://test.org')
 
                 pytest.raises(TypeError, resp.send_str, b's')
                 pytest.raises(TypeError, resp.send_bytes, 'b')
@@ -439,25 +412,24 @@ def test_reader_read_exception(ws_key, key_data, loop):
                 m_req.return_value = helpers.create_future(loop)
                 m_req.return_value.set_result(hresp)
                 WebSocketWriter.return_value = mock.Mock()
-                reader = mock.Mock()
-                hresp.connection.reader.set_parser.return_value = reader
 
-                resp = yield from aiohttp.ws_connect('http://test.org',
-                                                     loop=loop)
+                session = aiohttp.ClientSession(loop=loop)
+                resp = yield from session.ws_connect('http://test.org')
 
                 exc = ValueError()
-                reader.read.return_value = helpers.create_future(loop)
-                reader.read.return_value.set_exception(exc)
+                resp._reader.set_exception(exc)
 
                 msg = yield from resp.receive()
-                assert msg.type == aiohttp.MsgType.ERROR
+                assert msg.type == aiohttp.WSMsgType.ERROR
                 assert msg.type is msg.tp
                 assert resp.exception() is exc
 
+                session.close()
+
 
 @asyncio.coroutine
 def test_receive_runtime_err(loop):
-    resp = ClientWebSocketResponse(
+    resp = client.ClientWebSocketResponse(
         mock.Mock(), mock.Mock(), mock.Mock(), mock.Mock(), 10.0,
         True, True, loop)
     resp._waiting = True
@@ -481,10 +453,10 @@ def test_ws_connect_close_resp_on_err(loop, ws_key, key_data):
             m_req.return_value = helpers.create_future(loop)
             m_req.return_value.set_result(resp)
 
-            with pytest.raises(errors.WSServerHandshakeError):
-                yield from aiohttp.ws_connect('http://test.org',
-                                              protocols=('t1', 't2', 'chat'),
-                                              loop=loop)
+            with pytest.raises(client.WSServerHandshakeError):
+                yield from aiohttp.ClientSession(loop=loop).ws_connect(
+                    'http://test.org',
+                    protocols=('t1', 't2', 'chat'))
             resp.close.assert_called_with()
 
 
@@ -504,10 +476,9 @@ def test_ws_connect_non_overlapped_protocols(ws_key, loop, key_data):
             m_req.return_value = helpers.create_future(loop)
             m_req.return_value.set_result(resp)
 
-            res = yield from aiohttp.ws_connect(
+            res = yield from aiohttp.ClientSession(loop=loop).ws_connect(
                 'http://test.org',
-                protocols=('t1', 't2', 'chat'),
-                loop=loop)
+                protocols=('t1', 't2', 'chat'))
 
     assert res.protocol is None
 
@@ -529,10 +500,146 @@ def test_ws_connect_non_overlapped_protocols_2(ws_key, loop, key_data):
             m_req.return_value.set_result(resp)
 
             connector = aiohttp.TCPConnector(loop=loop, force_close=True)
-            res = yield from aiohttp.ws_connect(
+            res = yield from aiohttp.ClientSession(
+                connector=connector, loop=loop).ws_connect(
                 'http://test.org',
-                protocols=('t1', 't2', 'chat'),
-                connector=connector,
-                loop=loop)
+                protocols=('t1', 't2', 'chat'))
 
     assert res.protocol is None
+    del res
+
+
+@asyncio.coroutine
+def test_ws_connect_deflate(loop, ws_key, key_data):
+    resp = mock.Mock()
+    resp.status = 101
+    resp.headers = {
+        hdrs.UPGRADE: hdrs.WEBSOCKET,
+        hdrs.CONNECTION: hdrs.UPGRADE,
+        hdrs.SEC_WEBSOCKET_ACCEPT: ws_key,
+        hdrs.SEC_WEBSOCKET_EXTENSIONS: 'permessage-deflate',
+    }
+    with mock.patch('aiohttp.client.os') as m_os:
+        with mock.patch('aiohttp.client.ClientSession.get') as m_req:
+            m_os.urandom.return_value = key_data
+            m_req.return_value = helpers.create_future(loop)
+            m_req.return_value.set_result(resp)
+
+            res = yield from aiohttp.ClientSession(loop=loop).ws_connect(
+                'http://test.org', compress=15)
+
+    assert res.compress == 15
+    assert res.client_notakeover is False
+
+
+@asyncio.coroutine
+def test_ws_connect_deflate_server_not_support(loop, ws_key, key_data):
+    resp = mock.Mock()
+    resp.status = 101
+    resp.headers = {
+        hdrs.UPGRADE: hdrs.WEBSOCKET,
+        hdrs.CONNECTION: hdrs.UPGRADE,
+        hdrs.SEC_WEBSOCKET_ACCEPT: ws_key,
+    }
+    with mock.patch('aiohttp.client.os') as m_os:
+        with mock.patch('aiohttp.client.ClientSession.get') as m_req:
+            m_os.urandom.return_value = key_data
+            m_req.return_value = helpers.create_future(loop)
+            m_req.return_value.set_result(resp)
+
+            res = yield from aiohttp.ClientSession(loop=loop).ws_connect(
+                'http://test.org', compress=15)
+
+    assert res.compress == 0
+    assert res.client_notakeover is False
+
+
+@asyncio.coroutine
+def test_ws_connect_deflate_notakeover(loop, ws_key, key_data):
+    resp = mock.Mock()
+    resp.status = 101
+    resp.headers = {
+        hdrs.UPGRADE: hdrs.WEBSOCKET,
+        hdrs.CONNECTION: hdrs.UPGRADE,
+        hdrs.SEC_WEBSOCKET_ACCEPT: ws_key,
+        hdrs.SEC_WEBSOCKET_EXTENSIONS: 'permessage-deflate; '
+                                       'client_no_context_takeover',
+    }
+    with mock.patch('aiohttp.client.os') as m_os:
+        with mock.patch('aiohttp.client.ClientSession.get') as m_req:
+            m_os.urandom.return_value = key_data
+            m_req.return_value = helpers.create_future(loop)
+            m_req.return_value.set_result(resp)
+
+            res = yield from aiohttp.ClientSession(loop=loop).ws_connect(
+                'http://test.org', compress=15)
+
+    assert res.compress == 15
+    assert res.client_notakeover is True
+
+
+@asyncio.coroutine
+def test_ws_connect_deflate_client_wbits(loop, ws_key, key_data):
+    resp = mock.Mock()
+    resp.status = 101
+    resp.headers = {
+        hdrs.UPGRADE: hdrs.WEBSOCKET,
+        hdrs.CONNECTION: hdrs.UPGRADE,
+        hdrs.SEC_WEBSOCKET_ACCEPT: ws_key,
+        hdrs.SEC_WEBSOCKET_EXTENSIONS: 'permessage-deflate; '
+                                       'client_max_window_bits=10',
+    }
+    with mock.patch('aiohttp.client.os') as m_os:
+        with mock.patch('aiohttp.client.ClientSession.get') as m_req:
+            m_os.urandom.return_value = key_data
+            m_req.return_value = helpers.create_future(loop)
+            m_req.return_value.set_result(resp)
+
+            res = yield from aiohttp.ClientSession(loop=loop).ws_connect(
+                'http://test.org', compress=15)
+
+    assert res.compress == 10
+    assert res.client_notakeover is False
+
+
+@asyncio.coroutine
+def test_ws_connect_deflate_client_wbits_bad(loop, ws_key, key_data):
+    resp = mock.Mock()
+    resp.status = 101
+    resp.headers = {
+        hdrs.UPGRADE: hdrs.WEBSOCKET,
+        hdrs.CONNECTION: hdrs.UPGRADE,
+        hdrs.SEC_WEBSOCKET_ACCEPT: ws_key,
+        hdrs.SEC_WEBSOCKET_EXTENSIONS: 'permessage-deflate; '
+                                       'client_max_window_bits=6',
+    }
+    with mock.patch('aiohttp.client.os') as m_os:
+        with mock.patch('aiohttp.client.ClientSession.get') as m_req:
+            m_os.urandom.return_value = key_data
+            m_req.return_value = helpers.create_future(loop)
+            m_req.return_value.set_result(resp)
+
+            with pytest.raises(client.WSServerHandshakeError):
+                yield from aiohttp.ClientSession(loop=loop).ws_connect(
+                    'http://test.org', compress=15)
+
+
+@asyncio.coroutine
+def test_ws_connect_deflate_server_ext_bad(loop, ws_key, key_data):
+    resp = mock.Mock()
+    resp.status = 101
+    resp.headers = {
+        hdrs.UPGRADE: hdrs.WEBSOCKET,
+        hdrs.CONNECTION: hdrs.UPGRADE,
+        hdrs.SEC_WEBSOCKET_ACCEPT: ws_key,
+        hdrs.SEC_WEBSOCKET_EXTENSIONS: 'permessage-deflate; bad',
+    }
+    with mock.patch('aiohttp.client.os') as m_os:
+        with mock.patch('aiohttp.client.ClientSession.get') as m_req:
+            m_os.urandom.return_value = key_data
+            m_req.return_value = helpers.create_future(loop)
+            m_req.return_value.set_result(resp)
+
+            with pytest.raises(client.WSServerHandshakeError):
+                yield from aiohttp.ClientSession(loop=loop).ws_connect(
+                    'http://test.org', compress=15)
